"""Module containing common processing classes and functions."""

import os

MARKETPLACE_SLURM_OFFERING_TYPE = "Marketplace.Slurm"
WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES = int(
    os.environ.get("WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES", "5")
)
WALDUR_SITE_AGENT_REPORT_PERIOD_MINUTES = int(
    os.environ.get("WALDUR_SITE_AGENT_REPORT_PERIOD_MINUTES", "30")
)
WALDUR_SITE_AGENT_MEMBERSHIP_SYNC_PERIOD_MINUTES = int(
    os.environ.get("WALDUR_SITE_AGENT_MEMBERSHIP_SYNC_PERIOD_MINUTES", "5")
)



"""Utils for SLURM backend."""

from typing import Dict


def convert_slurm_units_to_waldur_ones(slurm_tres: Dict, units: Dict, to_int: bool = False) -> Dict:
    """Converts SLURM computing units to Waldur ones."""
    converted_units = {}

    for tres_name, value in units.items():
        converted_value = value
        converted_value = converted_value / slurm_tres[tres_name].get("unit_factor", 1)

        if to_int:
            converted_units[tres_name] = int(converted_value)
        else:
            converted_units[tres_name] = round(converted_value, 2)

    return converted_units



"""Module for order processing."""

from time import sleep

from waldur_site_agent.backends import logger
from waldur_site_agent.common import WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES, processors
from waldur_site_agent.common import structures as common_structures


def start(configuration: common_structures.WaldurAgentConfiguration) -> None:
    """Starts the main loop for offering processing."""
    waldur_offerings = configuration.waldur_offerings
    user_agent = configuration.waldur_user_agent
    while True:
        logger.info("Number of offerings to process: %s", len(waldur_offerings))
        for offering in waldur_offerings:
            try:
                if offering.mqtt_enabled:
                    logger.info(
                        "Skipping HTTP polling for the offering %s, because it uses mqtt feature",
                        offering.name,
                    )
                    continue

                processor = processors.OfferingOrderProcessor(offering, user_agent)
                processor.process_offering()
            except Exception as e:
                logger.exception("Unable to process the offering due to the error: %s", e)
        sleep(WALDUR_SITE_AGENT_ORDER_PROCESS_PERIOD_MINUTES * 60)



"""Module for abstract offering processor."""

from __future__ import annotations

import abc
import datetime
import traceback
from time import sleep
from typing import Dict, List, Optional, Set, Tuple

from waldur_client import (
    ComponentUsage,
    WaldurClient,
    is_uuid,
)

from waldur_site_agent.backends import BackendType, logger
from waldur_site_agent.backends import utils as backend_utils
from waldur_site_agent.backends.exceptions import BackendError
from waldur_site_agent.backends.structures import Resource
from waldur_site_agent.common import structures, utils


class OfferingBaseProcessor(abc.ABC):
    """Abstract class for an offering processing."""

    def __init__(self, offering: structures.Offering, user_agent: str = "") -> None:
        """Constructor."""
        self.offering: structures.Offering = offering
        self.waldur_rest_client: WaldurClient = WaldurClient(
            offering.api_url, offering.api_token, user_agent
        )
        self.resource_backend = utils.get_backend_for_offering(offering)
        if self.resource_backend.backend_type == BackendType.UNKNOWN.value:
            raise BackendError(f"Unable to create backend for {self.offering}")

        self._print_current_user()

        waldur_offering = self.waldur_rest_client.get_marketplace_provider_offering(
            self.offering.uuid
        )
        utils.extend_backend_components(self.offering, waldur_offering["components"])

    def _print_current_user(self) -> None:
        current_user = self.waldur_rest_client.get_current_user()
        utils.print_current_user(current_user)

    def _collect_waldur_resource_info(self, resource_data: dict) -> Resource:
        return Resource(
            name=resource_data["name"],
            backend_id=resource_data["backend_id"],
            marketplace_uuid=resource_data["uuid"],
            backend_type=self.offering.backend_type,
            restrict_member_access=resource_data.get("restrict_member_access", False),
            downscaled=resource_data.get("downscaled", False),
            paused=resource_data.get("paused", False),
            state=resource_data["state"],
        )

    @abc.abstractmethod
    def process_offering(self) -> None:
        """Pulls data form Mastermind using REST client and creates objects on the backend."""


class OfferingOrderProcessor(OfferingBaseProcessor):
    """Class for an offering processing.

    Processes related orders and creates necessary associations.
    """

    def process_offering(self) -> None:
        """Pulls data form Mastermind using REST client and creates objects on the backend."""
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )

        orders = self.waldur_rest_client.list_orders(
            {
                "offering_uuid": self.offering.uuid,
                "state": ["pending-provider", "executing"],
            }
        )

        if len(orders) == 0:
            logger.info("There are no pending or executing orders")
            return

        for order in orders:
            try:
                self.process_order_with_retries(order)
            except Exception as e:
                logger.exception(
                    "Error while processing offering order %s (%s), type %s, state %s: %s",
                    order["uuid"],
                    order["attributes"].get("name", "N/A"),
                    order["type"],
                    order["state"],
                    e,
                )

    def get_order_info(self, order_uuid: str) -> Optional[dict]:
        """Get order info from Waldur."""
        try:
            return self.waldur_rest_client.get_order(order_uuid)
        except Exception as e:
            logger.error("Failed to get order %s info: %s", order_uuid, e)
            return None

    def process_order_with_retries(
        self, order_info: dict, retry_count: int = 10, delay: int = 5
    ) -> None:
        """Process order with retries."""
        for attempt_number in range(retry_count):
            try:
                logger.info("Attempt %s of %s", attempt_number + 1, retry_count)
                order = self.get_order_info(order_info["uuid"])
                if order is None:
                    logger.error("Failed to get order %s info", order_info["uuid"])
                    return
                self.process_order(order)
                break
            except Exception as e:
                logger.exception(
                    "Error while processing order %s (%s), type %s, state %s: %s",
                    order_info["uuid"],
                    order_info["attributes"].get("name", "N/A"),
                    order_info["type"],
                    order["state"] if order is not None else order_info["state"],
                    e,
                )
                logger.info("Retrying order %s processing in %s seconds", order_info["uuid"], delay)
                sleep(delay)

        if attempt_number == retry_count - 1:
            logger.error(
                "Failed to process order %s after %s retries, skipping to the next one",
                order_info["uuid"],
                retry_count,
            )

    def process_order(self, order: dict) -> None:
        """Process a single order."""
        try:
            logger.info(
                "Processing order %s (%s) type %s, state %s",
                order["attributes"].get("name", "N/A"),
                order["uuid"],
                order["type"],
                order["state"],
            )

            if order["state"] == "executing":
                logger.info("Order is executing already, no need for approval")
            elif order["state"] == "pending-provider":
                logger.info("Approving the order")
                self.waldur_rest_client.marketplace_order_approve_by_provider(order["uuid"])
                logger.info("Refreshing the order")
                order = self.waldur_rest_client.get_order(order["uuid"])
            else:
                logger.warning(
                    "The order %s %s (%s) is in unexpected state %s, skipping processing",
                    order["type"],
                    order["resource_name"],
                    order["uuid"],
                    order["state"],
                )
                return

            order_is_done = False

            if order["type"] == "Create":
                order_is_done = self._process_create_order(order)

            if order["type"] == "Update":
                order_is_done = self._process_update_order(order)

            if order["type"] == "Terminate":
                order_is_done = self._process_terminate_order(order)

            # TODO: no need for update of orders for marketplace SLURM offerings
            if order_is_done:
                logger.info("Marking order as done")
                self.waldur_rest_client.marketplace_order_set_state_done(order["uuid"])

                logger.info("The order has been successfully processed")
            else:
                logger.warning("The order processing was not finished, skipping to the next one")

        except Exception as e:
            logger.exception(
                "Error while processing order %s: %s",
                order["uuid"],
                e,
            )
            self.waldur_rest_client.marketplace_order_set_state_erred(
                order["uuid"],
                error_message=str(e),
                error_traceback=traceback.format_exc(),
            )

    def _create_resource(
        self,
        waldur_resource: Dict,
    ) -> Resource | None:
        resource_uuid = waldur_resource["uuid"]
        resource_name = waldur_resource["name"]

        logger.info("Creating resource %s", resource_name)

        if not is_uuid(resource_uuid):
            logger.error("Unexpected resource UUID format, skipping the order")
            return None

        backend_resource = self.resource_backend.create_resource(waldur_resource)
        if backend_resource.backend_id == "":
            msg = f"Unable to create a backend resource for offering {self.offering}"
            raise BackendError(msg)

        logger.info("Updating resource metadata in Waldur")
        self.waldur_rest_client.marketplace_provider_resource_set_backend_id(
            resource_uuid, backend_resource.backend_id
        )

        return backend_resource

    def _add_users_to_resource(
        self,
        backend_resource: Resource,
    ) -> None:
        logger.info("Adding users to resource")
        logger.info("Fetching Waldur resource team")
        team = self.waldur_rest_client.marketplace_provider_resource_get_team(
            backend_resource.marketplace_uuid
        )
        user_uuids = {user["uuid"] for user in team}

        logger.info("Fetching Waldur offering users")
        offering_users_all = self.waldur_rest_client.list_remote_offering_users(
            {"offering_uuid": self.offering.uuid, "is_restricted": False}
        )
        offering_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users_all
            if offering_user["user_uuid"] in user_uuids and offering_user["username"] != ""
        }

        logger.info("Adding usernames to resource in backend")
        self.resource_backend.add_users_to_resource(
            backend_resource.backend_id,
            offering_usernames,
            homedir_umask=self.offering.backend_settings.get("homedir_umask", "0700"),
        )

    def _process_create_order(self, order: Dict) -> bool:
        # Wait until Waldur resource is created
        attempts = 0
        max_attempts = 4
        while "marketplace_resource_uuid" not in order:
            if attempts > max_attempts:
                logger.error("Order processing timed out")
                return False

            if order["state"] != "executing":
                logger.error("Order has unexpected state %s", order["state"])
                return False

            logger.info("Waiting for resource creation...")
            sleep(5)

            order = self.waldur_rest_client.get_order(order["uuid"])
            attempts += 1

        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(
            order["marketplace_resource_uuid"]
        )
        waldur_resource_info = self._collect_waldur_resource_info(waldur_resource)
        create_resource = True

        if waldur_resource_info.backend_id != "":
            logger.info(
                "Waldur resource backend id is not empty %s, checking backend resource data",
                waldur_resource_info.backend_id,
            )
            backend_resource = self.resource_backend.pull_resource(waldur_resource_info)
            if backend_resource is not None:
                logger.info(
                    "Resource %s (%s) is already created, skipping creation",
                    waldur_resource["name"],
                    waldur_resource["backend_id"],
                )
                create_resource = False

        if create_resource:
            waldur_resource["project_slug"] = order["project_slug"]
            waldur_resource["customer_slug"] = order["customer_slug"]

            backend_resource = self._create_resource(waldur_resource)
            if backend_resource is None:
                msg = "Unable to create a resource"
                raise BackendError(msg)

        if backend_resource is None:
            return False

        self._add_users_to_resource(
            backend_resource,
        )

        return True

    def _process_update_order(self, order: dict) -> bool:
        logger.info("Updating limits for %s", order["resource_name"])
        resource_uuid = order["marketplace_resource_uuid"]
        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(resource_uuid)

        resource_backend = utils.get_backend_for_offering(self.offering)
        if resource_backend is None:
            return False

        waldur_resource_backend_id = waldur_resource["backend_id"]

        new_limits = order["limits"]
        if not new_limits:
            logger.error(
                "Order %s (resource %s) with type" + "Update does not include new limits",
                order["uuid"],
                waldur_resource["name"],
            )

        if new_limits:
            resource_backend.set_resource_limits(waldur_resource_backend_id, new_limits)

        logger.info(
            "The limits for %s were updated successfully from %s to %s",
            waldur_resource["name"],
            order["attributes"]["old_limits"],
            new_limits,
        )
        return True

    def _process_terminate_order(self, order: dict) -> bool:
        logger.info("Terminating resource %s", order["resource_name"])
        resource_uuid = order["marketplace_resource_uuid"]

        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(resource_uuid)
        project_slug = order["project_slug"]

        resource_backend = utils.get_backend_for_offering(self.offering)
        if resource_backend is None:
            return False

        resource_backend.delete_resource(waldur_resource["backend_id"], project_slug=project_slug)

        logger.info("Allocation has been terminated successfully")
        return True


class OfferingMembershipProcessor(OfferingBaseProcessor):
    """Class for an offering processing.

    Processes related resources and reports membership data to Waldur.
    """

    def _get_waldur_resources(self, project_uuid: Optional[str] = None) -> List[Resource]:
        filters = {
            "offering_uuid": self.offering.uuid,
            "state": ["OK", utils.RESOURCE_ERRED_STATE],
            "field": [
                "backend_id",
                "uuid",
                "name",
                "resource_uuid",
                "offering_type",
                "restrict_member_access",
                "downscaled",
                "paused",
                "state",
                "limits",
            ],
        }

        if project_uuid is not None:
            filters["project_uuid"] = project_uuid

        waldur_resources = self.waldur_rest_client.filter_marketplace_provider_resources(filters)

        if len(waldur_resources) == 0:
            logger.info("No resources to process")
            return []

        return [
            self._collect_waldur_resource_info(resource_data)
            for resource_data in waldur_resources
            if resource_data["backend_id"]
        ]

    def process_resource_by_uuid(self, resource_uuid: str) -> None:
        """Processes resource status and membership data using resource UUID."""
        logger.info("Processing resource state and membership data, uuid: %s", resource_uuid)
        logger.info("Fetching resource from Waldur")
        waldur_resource = self.waldur_rest_client.get_marketplace_provider_resource(resource_uuid)
        resource_info = self._collect_waldur_resource_info(waldur_resource)
        logger.info(
            "Pulling resource %s (%s) from backend", resource_info.name, resource_info.backend_id
        )
        resource_report = self.resource_backend.pull_resources([resource_info])
        self._process_resources(resource_report)

    def process_offering(self) -> None:
        """Processes offering and reports resources usage to Waldur."""
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )

        waldur_resources_info = self._get_waldur_resources()
        resource_report = self.resource_backend.pull_resources(waldur_resources_info)

        self._process_resources(resource_report)

    def _get_user_offering_users(self, user_uuid: str) -> List[dict]:
        return self.waldur_rest_client.list_remote_offering_users(
            {
                "offering_uuid": self.offering.uuid,
                "user_uuid": user_uuid,
                "is_restricted": False,
            }
        )

    def process_user_role_changed(self, user_uuid: str, project_uuid: str, granted: bool) -> None:
        """Process event of user role changing."""
        offering_users = self._get_user_offering_users(user_uuid)
        if len(offering_users) == 0:
            logger.info(
                "User %s is not linked to the offering %s (%s)",
                user_uuid,
                self.offering.name,
                self.offering.uuid,
            )
            return

        username = offering_users[0]["username"]
        logger.info("Using offering user with username %s", username)
        if not username:
            logger.warning("Username is blank, skipping processing")
            return

        resources = self._get_waldur_resources(project_uuid=project_uuid)
        resource_report = self.resource_backend.pull_resources(resources)

        for resource in resource_report.values():
            try:
                if granted:
                    if resource.restrict_member_access:
                        logger.info("The resource is restricted, skipping new role.")
                        continue
                    self.resource_backend.add_user(resource.backend_id, username)
                else:
                    self.resource_backend.remove_user(resource.backend_id, username)
            except Exception as exc:
                logger.error(
                    "Unable to add user %s to the resource %s, error: %s",
                    username,
                    resource.backend_id,
                    exc,
                )

    def _get_waldur_offering_users(self) -> List[Dict]:
        logger.info("Fetching Waldur offering users")
        return self.waldur_rest_client.list_remote_offering_users(
            {
                "offering_uuid": self.offering.uuid,
                "is_restricted": False,
            }
        )

    def _get_waldur_resource_team(self, resource: Resource) -> List[Dict]:
        logger.info("Fetching Waldur resource team")
        return self.waldur_rest_client.marketplace_provider_resource_get_team(
            resource.marketplace_uuid
        )

    def _get_resource_usernames(self, resource: Resource) -> Tuple[Set[str], Set[str], Set[str]]:
        logger.info("Fetching new, existing and stale resource users")
        usernames = resource.users
        local_usernames = set(usernames)
        logger.info("The usernames from the backend: %s", ", ".join(local_usernames))

        # Offering users sync
        # The service fetches offering users from Waldur and pushes them to the cluster
        # If an offering user is not in the team anymore, it will be removed from the backend
        team = self._get_waldur_resource_team(resource)
        team_user_uuids = {user["uuid"] for user in team}

        offering_users = self._get_waldur_offering_users()
        resource_offering_usernames = {
            offering_user["username"]
            for offering_user in offering_users
            if offering_user["user_uuid"] in team_user_uuids
        }
        logger.info("Resource offering usernames: %s", ", ".join(resource_offering_usernames))

        existing_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users
            if offering_user["username"] in local_usernames
            and offering_user["user_uuid"] in team_user_uuids
        }
        logger.info("Resource existing usernames: %s", ", ".join(existing_usernames))

        new_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users
            if offering_user["username"] not in local_usernames
            and offering_user["user_uuid"] in team_user_uuids
        }
        logger.info("Resource new usernames: %s", ", ".join(new_usernames))

        stale_usernames: Set[str] = {
            offering_user["username"]
            for offering_user in offering_users
            if offering_user["username"] in local_usernames
            and offering_user["user_uuid"] not in team_user_uuids
        }
        logger.info("Resource stale usernames: %s", ", ".join(stale_usernames))

        return existing_usernames, stale_usernames, new_usernames

    def _sync_resource_users(
        self,
        resource: Resource,
    ) -> Set[str]:
        """Sync users for the resource between Waldur and the site.

        return: the actual resource usernames (existing + added)
        """
        logger.info("Syncing user list for resource %s", resource.name)
        existing_usernames, stale_usernames, new_usernames = self._get_resource_usernames(resource)

        if resource.restrict_member_access:
            # The idea is to remove the existing associations in both sides
            # and avoid creation of new associations
            logger.info(
                "Resource is restricted for members, removing all the existing associations"
            )

            self.resource_backend.remove_users_from_account(resource.backend_id, existing_usernames)
            return set()

        added_usernames = self.resource_backend.add_users_to_resource(
            resource.backend_id,
            new_usernames,
            homedir_umask=self.offering.backend_settings.get("homedir_umask", "0700"),
        )

        self.resource_backend.remove_users_from_account(
            resource.backend_id,
            stale_usernames,
        )

        return existing_usernames | added_usernames

    def _sync_resource_status(self, resource: Resource) -> None:
        """Syncs resource status between Waldur and the backend."""
        logger.info(
            "Syncing resource status for resource %s (%s)", resource.name, resource.backend_id
        )
        if resource.paused:
            logger.info("Resource pausing is requested, processing it")
            pausing_done = self.resource_backend.pause_resource(resource.backend_id)
            if pausing_done:
                logger.info("Pausing is successfully completed")
            else:
                logger.warning("Pausing is not done")
        elif resource.downscaled:
            logger.info("Resource downscaling is requested, processing it")
            downscaling_done = self.resource_backend.downscale_resource(resource.backend_id)
            if downscaling_done:
                logger.info("Downscaling is successfully completed")
            else:
                logger.warning("Downscaling is not done")
        else:
            logger.info(
                "The resource is not downscaled or paused, resetting the QoS to the default one"
            )
            restoring_done = self.resource_backend.restore_resource(resource.backend_id)
            if restoring_done:
                logger.info("Restoring is successfully completed")
            else:
                logger.info("Restoring is skipped")

        resource_metadata = self.resource_backend.get_resource_metadata(resource.backend_id)
        self.waldur_rest_client.marketplace_provider_resource_set_backend_metadata(
            resource.marketplace_uuid, resource_metadata
        )

    def _sync_resource_limits(self, resource: Resource) -> None:
        """Syncs resource limits between Waldur and the backend."""
        logger.info(
            "Syncing resource limits for resource %s (%s)", resource.name, resource.backend_id
        )
        waldur_limits = resource.limits
        backend_limits = self.resource_backend.get_resource_limits(resource.backend_id)

        if len(backend_limits) == 0:
            logger.warning("No limits are found in the backend")
            return

        if backend_limits == waldur_limits:
            logger.info("The limits are already in sync, skipping")
            return
        # For now, we report all the limits
        logger.info("Reporting the limits to Waldur: %s", backend_limits)
        self.waldur_rest_client.marketplace_provider_resource_set_limits(
            resource.marketplace_uuid, backend_limits
        )

    # TODO: adapt for RabbitMQ-based processing
    # introduce new event and add support for the event in the agent
    def _sync_resource_user_limits(
        self, resource: Resource, usernames: Optional[Set[str]] = None
    ) -> None:
        logger.info(
            "Synching resource user limits for resource %s (%s)", resource.name, resource.backend_id
        )
        if resource.restrict_member_access:
            logger.info("Resource is restricted for members, skipping user limits setup")
            return

        if usernames is None:
            existing_usernames, _, _ = self._get_resource_usernames(resource)
            usernames = existing_usernames

        backend_user_limits = self.resource_backend.get_resource_user_limits(resource.backend_id)

        for username in usernames:
            try:
                logger.info(
                    "Fetching user usage limits for %s, resource %s",
                    username,
                    resource.marketplace_uuid,
                )
                user_limits = self.waldur_rest_client.list_component_user_usage_limits(
                    {"resource_uuid": resource.marketplace_uuid, "username": username}
                )
                if len(user_limits) == 0:
                    existing_user_limits = backend_user_limits.get(username)
                    logger.info("The limits for user %s are not defined in Waldur")
                    if existing_user_limits is None:
                        continue
                    logger.info("Unsetting the existing limits %s", existing_user_limits)
                    user_component_limits = {}
                else:
                    user_component_limits = {
                        user_limit["component_type"]: int(float(user_limit["limit"]))
                        for user_limit in user_limits
                    }
                self.resource_backend.set_resource_user_limits(
                    resource.backend_id, username, user_component_limits
                )
            except Exception as exc:
                logger.error(
                    "Unable to set user %s limits for resource %s (%s), reason: %s",
                    username,
                    resource.name,
                    resource.backend_id,
                    exc,
                )

    def _process_resources(
        self,
        resource_report: Dict[str, Resource],
    ) -> None:
        """Sync status and membership data for the resource."""
        for backend_resource in resource_report.values():
            try:
                resource_usernames = self._sync_resource_users(backend_resource)
                self._sync_resource_status(backend_resource)
                self._sync_resource_limits(backend_resource)
                self._sync_resource_user_limits(backend_resource, resource_usernames)

                logger.info(
                    "Refreshing resource %s (%s) last sync",
                    backend_resource.name,
                    backend_resource.backend_id,
                )
                self.waldur_rest_client.marketplace_provider_resource_refresh_last_sync(
                    backend_resource.marketplace_uuid
                )
                if backend_resource.state == utils.RESOURCE_ERRED_STATE:
                    logger.info(
                        "Setting resource %s (%s) state to OK",
                        backend_resource.name,
                        backend_resource.backend_id,
                    )
                    self.waldur_rest_client.marketplace_provider_resource_set_as_ok(
                        backend_resource.marketplace_uuid
                    )
            except Exception as e:
                logger.exception(
                    "Error while processing allocation %s: %s",
                    backend_resource.backend_id,
                    e,
                )
                error_traceback = traceback.format_exc()
                utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [backend_resource],
                    error_details={
                        "error_message": str(e),
                        "error_traceback": error_traceback,
                    },
                )


class OfferingReportProcessor(OfferingBaseProcessor):
    """Class for an offering processing.

    Processes related resource and reports computing data to Waldur.
    """

    def process_offering(self) -> None:
        """Processes offering and reports resources usage to Waldur."""
        logger.info(
            "Processing offering %s (%s)",
            self.offering.name,
            self.offering.uuid,
        )

        waldur_offering = self.waldur_rest_client.get_marketplace_provider_offering(
            self.offering.uuid
        )

        waldur_resources = self.waldur_rest_client.filter_marketplace_provider_resources(
            {
                "offering_uuid": self.offering.uuid,
                "state": ["OK", utils.RESOURCE_ERRED_STATE],
                "field": ["backend_id", "uuid", "name", "offering_type", "state"],
            }
        )

        if len(waldur_resources) == 0:
            logger.info("No resources to process")
            return

        waldur_resources_info = [
            Resource(
                name=resource_data["name"],
                backend_id=resource_data["backend_id"],
                marketplace_uuid=resource_data["uuid"],
                backend_type=self.offering.backend_type,
                state=resource_data["state"],
            )
            for resource_data in waldur_resources
            if resource_data["backend_id"]
        ]

        for waldur_resource in waldur_resources_info:
            try:
                self._process_resource_with_retries(waldur_resource, waldur_offering)
            except Exception as e:
                logger.exception(
                    "Error while processing allocation %s: %s",
                    waldur_resource.backend_id,
                    e,
                )
                error_traceback = traceback.format_exc()
                utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [waldur_resource],
                    error_details={
                        "error_message": str(e),
                        "error_traceback": error_traceback,
                    },
                )

    def _process_resource_with_retries(
        self,
        waldur_resource: Resource,
        waldur_offering: Dict,
        retry_count: int = 10,
        delay: int = 5,
    ) -> None:
        for attempt_number in range(retry_count):
            try:
                logger.info(
                    "Attempt %s of %s, processing resource usage %s (%s)",
                    attempt_number + 1,
                    retry_count,
                    waldur_resource.name,
                    waldur_resource.backend_id,
                )
                self._process_resource(waldur_resource, waldur_offering)
                break
            except Exception as e:
                logger.warning(
                    "Error while processing resource %s (%s): %s",
                    waldur_resource.name,
                    waldur_resource.backend_id,
                    e,
                )
                logger.info(
                    "Retrying resource usage %s processing in %s seconds",
                    waldur_resource.backend_id,
                    delay,
                )
                if attempt_number == retry_count - 1:
                    # If last attempt failed, raise the exception
                    logger.warning(
                        "Failed to process resource usage %s after %s retries,"
                        "skipping to the next resource",
                        waldur_resource.backend_id,
                        retry_count,
                    )
                    raise
                # If not last attempt, wait and retry
                sleep(delay)

    def _submit_total_usage_for_resource(
        self,
        backend_resource: Resource,
        total_usage: Dict[str, float],
        waldur_components: List[Dict],
    ) -> None:
        """Reports total usage for a backend resource to Waldur."""
        logger.info("Setting usages for %s: %s", backend_resource.backend_id, total_usage)
        resource_uuid = backend_resource.marketplace_uuid

        component_types = [component["type"] for component in waldur_components]
        missing_components = set(total_usage) - set(component_types)

        if missing_components:
            logger.warning(
                "The following components are not found in Waldur: %s",
                ", ".join(missing_components),
            )

        usage_objects = [
            ComponentUsage(type=component, amount=amount)
            for component, amount in total_usage.items()
            if component in component_types
        ]
        self.waldur_rest_client.create_component_usages(
            resource_uuid=resource_uuid, usages=usage_objects
        )

    def _submit_user_usage_for_resource(
        self,
        username: str,
        user_usage: Dict[str, float],
        waldur_component_usages: List[Dict],
    ) -> None:
        """Reports per-user usage for a backend resource to Waldur."""
        logger.info("Setting usages for %s", username)
        component_usage_types = [
            component_usage["type"] for component_usage in waldur_component_usages
        ]
        missing_components = set(user_usage) - set(component_usage_types)

        if missing_components:
            logger.warning(
                "The following components are not found in Waldur: %s",
                ", ".join(missing_components),
            )

        offering_users = self.waldur_rest_client.list_remote_offering_users(
            {"username": username, "query": self.offering.uuid}
        )
        offering_user_uuid = None

        if len(offering_users) > 0:
            offering_user_uuid = offering_users[0]["uuid"]

        for component_usage in waldur_component_usages:
            component_type = component_usage["type"]
            usage = user_usage.get(component_type)
            if usage is None:
                logger.warning(
                    "No usage for Waldur component %s is found in SLURM user usage report",
                    component_type,
                )
                continue
            logger.info(
                "Submitting usage for username %s: %s -> %s",
                username,
                component_type,
                usage,
            )
            self.waldur_rest_client.create_component_user_usage(
                component_usage["uuid"], usage, username, offering_user_uuid
            )

    def _process_resource(
        self,
        waldur_resource: Resource,
        waldur_offering: Dict,
    ) -> None:
        """Processes usage report for the resource."""
        month_start = backend_utils.month_start(datetime.datetime.now()).date()
        resource_backend_id = waldur_resource.backend_id
        logger.info("Pulling resource %s (%s)", waldur_resource.name, resource_backend_id)
        backend_resource = self.resource_backend.pull_resource(waldur_resource)
        if backend_resource is None:
            logger.info("The resource %s is missing in backend", resource_backend_id)
            if waldur_resource.state != utils.RESOURCE_ERRED_STATE:
                logger.info("Marking resource %s as erred in Waldur", resource_backend_id)
                utils.mark_waldur_resources_as_erred(
                    self.waldur_rest_client,
                    [waldur_resource],
                    {
                        "error_message": f"The resource {resource_backend_id} "
                        "is missing on the backend"
                    },
                )
            return

        # Invalidate cache of Waldur resource
        logger.info(
            "Fetching Waldur resource data for %s (%s)", waldur_resource.name, resource_backend_id
        )
        waldur_resource_data = self.waldur_rest_client.get_marketplace_provider_resource(
            waldur_resource.marketplace_uuid
        )
        waldur_resource = self._collect_waldur_resource_info(waldur_resource_data)
        waldur_resource.usage = backend_resource.usage
        waldur_resource.users = backend_resource.users
        waldur_resource.limits = backend_resource.limits

        if waldur_resource.state not in ["OK", utils.RESOURCE_ERRED_STATE]:
            logger.error(
                "Waldur resource %s (%s) has incorrect state %s, skipping processing",
                waldur_resource.name,
                waldur_resource.backend_id,
                waldur_resource.state,
            )
            return
        # Set resource state OK if it is erred
        if waldur_resource.state == utils.RESOURCE_ERRED_STATE:
            self.waldur_rest_client.marketplace_provider_resource_set_as_ok(
                waldur_resource.marketplace_uuid
            )

        usages: Dict[str, Dict[str, float]] = waldur_resource.usage

        # Submit usage
        total_usage = usages.pop("TOTAL_ACCOUNT_USAGE")
        self._submit_total_usage_for_resource(
            waldur_resource,
            total_usage,
            waldur_offering["components"],
        )

        # Skip the following actions if the dict is empty
        if not usages:
            return

        waldur_component_usages = self.waldur_rest_client.list_component_usages(
            waldur_resource.marketplace_uuid, date_after=month_start
        )

        logger.info("Setting per-user usages")
        for username, user_usage in usages.items():
            self._submit_user_usage_for_resource(username, user_usage, waldur_component_usages)



"""Functions for management of MQTT signals."""

from __future__ import annotations

import json
import signal
import sys
import types
from contextlib import contextmanager
from typing import Generator, List, Union

import paho.mqtt.client as mqtt

from waldur_site_agent.backends import logger
from waldur_site_agent.common import processors as common_processors
from waldur_site_agent.common import structures as common_structures
from waldur_site_agent.event_processing.event_subscription_manager import EventSubscriptionManager
from waldur_site_agent.event_processing.structures import (
    MqttConsumer,
    MqttConsumersMap,
    OrderMessage,
    ResourceMessage,
    UserData,
    UserRoleMessage,
)


def on_connect(
    client: mqtt.Client,
    userdata: UserData,
    flags: mqtt.ConnectFlags,
    reason_code: mqtt.ReasonCode,
    properties: Union[mqtt.Properties, None],
) -> None:
    """Order-processing handler for MQTT connection event."""
    del flags, properties
    logger.debug("Consumer connected with result code %s", reason_code)
    offering_uuid = userdata["offering"].uuid
    if reason_code.is_failure:
        logger.error("Consumer connection error (%s): %s", offering_uuid, reason_code.getName())
    else:
        event_subscription_uuid = userdata["event_subscription"]["uuid"]
        topic_postfix = userdata["topic_postfix"]
        topic_name = (
            f"subscription/{event_subscription_uuid}/offering/{offering_uuid}/{topic_postfix}"
        )
        logger.debug("Subscribing to the topic %s", topic_name)
        client.subscribe(topic_name)


def setup_offering_subscriptions(
    waldur_offering: common_structures.Offering, waldur_user_agent: str
) -> List[MqttConsumer]:
    """Set up MQTT subscriptions for the specified offering."""
    object_type_to_handler = {
        "order": on_order_message,
        "user_role": on_user_role_message,
        "resource": on_resource_message,
    }

    event_subscriptions: List[MqttConsumer] = []
    for object_type in ["order", "user_role", "resource"]:
        on_message_handler = object_type_to_handler[object_type]
        event_subscription_manager = EventSubscriptionManager(
            waldur_offering, on_connect, on_message_handler, waldur_user_agent, object_type
        )
        event_subscription = event_subscription_manager.create_event_subscription()
        if event_subscription is None:
            logger.error(
                "Failed to create event subscription for the offering %s (%s), object type %s",
                waldur_offering.name,
                waldur_offering.uuid,
                object_type,
            )
            continue
        consumer = event_subscription_manager.start_mqtt_consumer(event_subscription)
        if consumer is None:
            logger.error(
                "Failed to start mqtt consumer for the offering %s (%s), object type %s",
                waldur_offering.name,
                waldur_offering.uuid,
                object_type,
            )
            event_subscription_manager.delete_event_subscription(event_subscription)
            continue

        event_subscriptions.append((consumer, event_subscription, waldur_offering))

    return event_subscriptions


def start_mqtt_consumers(
    waldur_offerings: List[common_structures.Offering],
    waldur_user_agent: str,
) -> MqttConsumersMap:
    """Start multiple MQTT consumers."""
    mqtt_consumers_map: MqttConsumersMap = {}
    for waldur_offering in waldur_offerings:
        if not waldur_offering.mqtt_enabled:
            logger.info("MQTT feature is disabled for the offering")
            continue

        event_subscriptions = setup_offering_subscriptions(waldur_offering, waldur_user_agent)
        if event_subscriptions:
            mqtt_consumers_map[(waldur_offering.name, waldur_offering.uuid)] = event_subscriptions

    return mqtt_consumers_map


def stop_mqtt_consumers(
    mqtt_consumers_map: MqttConsumersMap,
) -> None:
    """Stop mqtt consumers and delete event subscriptions."""
    for (offering_name, offering_uuid), subscriptions in mqtt_consumers_map.items():
        logger.info("Stopping subscriptions for %s (%s)", offering_name, offering_uuid)
        for (
            mqttc,
            event_subscription,
            offering,
        ) in subscriptions:
            try:
                event_subscription_manager = EventSubscriptionManager(
                    offering,
                )
                logger.info(
                    "Stopping MQTT consumer for %s (%s), observable object type: %s",
                    offering_name,
                    offering_uuid,
                    event_subscription["observable_objects"][0]["object_type"],
                )
                event_subscription_manager.stop_mqtt_consumer(mqttc)
                logger.info("Deleting event subscription for %s (%s)", offering_name, offering_uuid)
                event_subscription_manager.delete_event_subscription(event_subscription)
            except Exception as exc:
                logger.exception("Unable to stop the consumer, reason: %s", exc)


@contextmanager
def signal_handling(
    mqtt_consumers_map: MqttConsumersMap,
) -> Generator[None, None, None]:
    """Context manager for handling signals gracefully."""

    def signal_handler(signum: int, _: types.FrameType | None) -> None:
        signal_name = signal.Signals(signum).name
        logger.info("Received %s signal. Shutting down gracefully...", signal_name)
        stop_mqtt_consumers(mqtt_consumers_map)
        sys.exit(0)

    # Register signal handlers
    signals = (
        signal.SIGTERM,
        signal.SIGINT,
        signal.SIGTSTP,
        signal.SIGQUIT,
    )
    original_handlers = {}

    try:
        logger.info("Setting up signal handlers")
        # Save original handlers and set new ones
        for sig in signals:
            original_handlers[sig] = signal.signal(sig, signal_handler)
        logger.info("Signal handlers set")
        yield
    finally:
        # Restore original handlers
        for sig, handler in original_handlers.items():
            signal.signal(sig, handler)


def on_order_message(client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage) -> None:
    """Order-processing handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: OrderMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]

    order_uuid = message["order_uuid"]
    try:
        processor = common_processors.OfferingOrderProcessor(offering, user_agent)
        order = processor.get_order_info(order_uuid)
        if order is None:
            logger.error("Failed to process order %s", order_uuid)
            return
        processor.process_order_with_retries(order)
    except Exception as e:
        logger.error("Failed to process order %s: %s", order_uuid, e)


def on_user_role_message(client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage) -> None:
    """Membership sync handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: UserRoleMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]
    user_uuid = message["user_uuid"]
    user_username = message["user_username"]
    project_name = message["project_name"]
    project_uuid = message["project_uuid"]
    role_granted = message["granted"]

    try:
        processor = common_processors.OfferingMembershipProcessor(offering, user_agent)
        logger.info(
            "Processing %s (%s) user role changed event in project %s, granted: %s",
            user_username,
            user_uuid,
            project_name,
            role_granted,
        )

        processor.process_user_role_changed(user_uuid, project_uuid, role_granted)
    except Exception as e:
        logger.error(
            "Failed to process user %s (%s) role change in project %s (%s) (granted: %s): %s",
            user_username,
            user_uuid,
            project_name,
            project_uuid,
            role_granted,
            e,
        )


def on_resource_message(client: mqtt.Client, userdata: UserData, msg: mqtt.MQTTMessage) -> None:
    """Resource update handler for MQTT message event."""
    del client
    message_text = msg.payload.decode("utf-8")
    message: ResourceMessage = json.loads(message_text)
    logger.info("Received message: %s on topic %s", message, msg.topic)
    offering = userdata["offering"]
    user_agent = userdata["user_agent"]
    resource_uuid = message["resource_uuid"]

    try:
        processor = common_processors.OfferingMembershipProcessor(offering, user_agent)
        processor.process_resource_by_uuid(resource_uuid)
    except Exception as e:
        logger.error("Failed to process resource %s: %s", resource_uuid, e)


def run_initial_offering_processing(
    waldur_offerings: List[common_structures.Offering], user_agent: str = ""
) -> None:
    """Runs processing of offerings with MQTT feature enabled."""
    logger.info("Processing offerings with MQTT feature enabled")
    for offering in waldur_offerings:
        try:
            if not offering.mqtt_enabled:
                continue

            process_offering(offering, user_agent)
        except Exception as e:
            logger.exception("Error occurred during initial offering process: %s", e)


def process_offering(offering: common_structures.Offering, user_agent: str = "") -> None:
    """Processes the specified offering."""
    logger.info("Processing offering %s (%s)", offering.name, offering.uuid)

    order_processor = common_processors.OfferingOrderProcessor(offering, user_agent)
    logger.info("Running offering order process")
    order_processor.process_offering()

    membership_processor = common_processors.OfferingMembershipProcessor(offering, user_agent)
    logger.info("Running offering membership process")
    membership_processor.process_offering()



"""Generic backend classes."""

from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Set, Tuple

from waldur_site_agent.backends import BackendType, logger, structures, utils
from waldur_site_agent.backends.base import BaseClient, UnknownClient
from waldur_site_agent.backends.exceptions import BackendError

UNKNOWN_BACKEND_TYPE = "unknown"


class BaseBackend(ABC):
    """Abstract backend class."""

    def __init__(self, backend_settings: Dict, backend_components: Dict[str, Dict]) -> None:
        """Init backend info."""
        self.backend_type = "abstract"
        self.backend_settings = backend_settings
        self.backend_components = backend_components
        self.client: BaseClient = UnknownClient()

    @abstractmethod
    def ping(self, raise_exception: bool = False) -> bool:
        """Check if backend is online."""

    @abstractmethod
    def list_components(self) -> List[str]:
        """Return a list of computing components on the backend."""

    def pull_resources(
        self, resources_info: List[structures.Resource]
    ) -> Dict[str, structures.Resource]:
        """Pull data of resources available in the backend."""
        report = {}
        for resource_info in resources_info:
            backend_id = resource_info.backend_id
            try:
                resource = self.pull_resource(resource_info)
                if resource is not None:
                    report[backend_id] = resource
            except Exception as e:
                logger.exception("Error while pulling allocation [%s]: %s", backend_id, e)
        return report

    def pull_resource(self, resource_info: structures.Resource) -> Optional[structures.Resource]:
        """Pull resource from backend."""
        try:
            backend_id = resource_info.backend_id
            backend_resource = self._pull_allocation(backend_id)
            if backend_resource is None:
                return None
        except Exception as e:
            logger.exception("Error while pulling resource [%s]: %s", backend_id, e)
            return None
        else:
            backend_resource.name = resource_info.name
            backend_resource.marketplace_uuid = resource_info.marketplace_uuid
            backend_resource.restrict_member_access = resource_info.restrict_member_access
            backend_resource.downscaled = resource_info.downscaled
            backend_resource.paused = resource_info.paused
            backend_resource.state = resource_info.state
            return backend_resource

    def _pull_allocation(self, resource_backend_id: str) -> Optional[structures.Resource]:
        """Pull allocation data from the backend."""
        account = resource_backend_id
        logger.info("Pulling allocation %s", account)
        account_info = self.client.get_account(account)

        if account_info is None:
            logger.warning("There is no %s account in the backend", account)
            return None

        users = self.client.list_account_users(account)

        report = self._get_usage_report([account])
        usage = report.get(account)

        if usage is None:
            empty_usage = {tres: 0 for tres in self.backend_components}
            usage = {"TOTAL_ACCOUNT_USAGE": empty_usage}

        return structures.Resource(
            name="",
            marketplace_uuid="",
            backend_id=account,
            limits={},
            users=users,
            usage=usage,
            backend_type=self.backend_type,
        )

    @abstractmethod
    def _get_usage_report(self, accounts: List[str]) -> Dict:
        """Collect usage report for the specified accounts."""

    def _pre_delete_account_actions(self, account: str) -> None:
        """Perform actions before deleting the account."""
        del account

    def delete_resource(self, resource_backend_id: str, **kwargs: str) -> None:
        """Delete resource from the backend."""
        account = resource_backend_id

        if not account.strip():
            logger.warning("Empty backend_id for allocation, skipping deletion")
            return

        if self.client.get_account(account) is None:
            logger.warning("No account %s is in %s", account, self.backend_type)
            return

        self._pre_delete_account_actions(account)
        self._delete_account_safely(account)

        if "project_slug" in kwargs:
            project_account = self._get_project_name(kwargs["project_slug"])
            if (
                len(
                    [
                        account
                        for account in self.client.list_accounts()
                        if account.organization == project_account
                        and account.name != project_account
                    ]
                )
                == 0
            ):
                self._delete_account_safely(project_account)

    def _delete_account_safely(self, account: str) -> None:
        if self.client.get_account(account):
            logger.info("Deleting the account %s from %s", account, self.backend_type)
            self.client.delete_account(account)
        else:
            logger.warning("No account %s is in %s", account, self.backend_type)

    def _create_account(
        self,
        account_name: str,
        account_description: str,
        account_organization: str,
        account_parent_name: Optional[str] = None,
    ) -> bool:
        logger.info(
            "Creating SLURM account %s (backend id = %s)",
            account_description,
            account_name,
        )
        if self.client.get_account(account_name) is None:
            self.client.create_account(
                name=account_name,
                description=account_description,
                organization=account_organization,
                parent_name=account_parent_name,
            )
            return True
        logger.info("The account %s already exists in the cluster", account_name)
        return False

    def create_resource(self, waldur_resource: Dict) -> structures.Resource:
        """Create resource on the backend.

        Creates necessary accounts hierarchy and sets up resource limits.
        """
        logger.info("Creating account in the backend")

        # Setup accounts hierarchy
        self._setup_accounts_hierarchy(waldur_resource)

        # Create allocation account
        allocation_account = self._create_allocation_account(waldur_resource)

        # Setup limits
        self._setup_resource_limits(allocation_account, waldur_resource)

        return structures.Resource(
            backend_type=self.backend_type,
            name=waldur_resource["name"],
            marketplace_uuid=waldur_resource["uuid"],
            backend_id=allocation_account,
            limits=self._collect_limits(waldur_resource)[1],
        )

    def _setup_accounts_hierarchy(self, waldur_resource: Dict) -> None:
        """Setup customer and project accounts hierarchy."""
        project_account = self._get_project_name(waldur_resource["project_slug"])

        # Setup customer account if using SLURM
        customer_account = None
        if self.backend_type == BackendType.SLURM.value:
            customer_account = self._get_customer_name(waldur_resource["customer_slug"])
            self._create_account(
                customer_account, waldur_resource["customer_name"], customer_account
            )

        # Create project account
        self._create_account(
            project_account, waldur_resource["project_name"], project_account, customer_account
        )

    def _create_allocation_account(self, waldur_resource: Dict) -> str:
        """Create allocation account with retry logic for name generation."""
        project_account = self._get_project_name(waldur_resource["project_slug"])

        # Determine allocation name generation strategy
        use_project_slug = (
            waldur_resource["offering_plugin_options"].get("account_name_generation_policy")
            == "project_slug"
        )

        allocation_name = (
            waldur_resource["project_slug"] if use_project_slug else waldur_resource["slug"]
        )
        max_retries = 10 if use_project_slug else 1

        # Try creating account with generated names
        for retry in range(max_retries):
            allocation_account = self._get_allocation_name(allocation_name)
            if self._create_account(
                allocation_account, waldur_resource["name"], project_account, project_account
            ):
                return allocation_account

            if use_project_slug:
                allocation_name = f"{waldur_resource['project_slug']}-{retry}"

        raise BackendError(
            f"Unable to create an allocation: {allocation_account} already exists in the cluster"
        )

    def _setup_resource_limits(self, allocation_account: str, waldur_resource: Dict) -> None:
        """Setup resource limits for the allocation account."""
        allocation_limits, _ = self._collect_limits(waldur_resource)

        if not allocation_limits:
            logger.info("Skipping setting of limits")
            return

        # Convert limits for logging
        converted_limits = {
            key: value // self.backend_components[key].get("unit_factor", 1)
            for key, value in allocation_limits.items()
        }

        limits_str = utils.prettify_limits(converted_limits, self.backend_components)
        logger.info("Setting allocation limits to: \n%s", limits_str)
        self.client.set_resource_limits(allocation_account, allocation_limits)

    @abstractmethod
    def downscale_resource(self, account: str) -> bool:
        """Downscale the account on the backend."""

    @abstractmethod
    def pause_resource(self, account: str) -> bool:
        """Pause the account on the backend."""

    @abstractmethod
    def restore_resource(self, account: str) -> bool:
        """Restore the account after downscaling or pausing."""

    @abstractmethod
    def get_resource_metadata(self, _: str) -> dict:
        """Get backend-specific resource metadata."""

    @abstractmethod
    def _collect_limits(
        self, waldur_resource: Dict[str, Dict]
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """Collect limits for backend and waldur separately."""

    def add_users_to_resource(
        self, resource_backend_id: str, user_ids: Set[str], **kwargs: dict
    ) -> Set[str]:
        """Add specified users to the resource on the backend."""
        del kwargs
        logger.info(
            "Adding users to account %s on backend: %s", resource_backend_id, " ,".join(user_ids)
        )
        added_users = set()
        for username in user_ids:
            try:
                succeeded = self.add_user(resource_backend_id, username)
                if succeeded:
                    added_users.add(username)
            except BackendError as e:
                logger.exception(
                    "Unable to add user %s to account %s, details: %s",
                    username,
                    resource_backend_id,
                    e,
                )

        return added_users

    def add_user(self, account: str, username: str) -> bool:
        """Add association between user and backend account if it doesn't exists."""
        if not account.strip():
            message = "Empty backend_id for allocation"
            raise BackendError(message)

        logger.info("Adding user %s to account %s", username, account)
        if not username:
            logger.warning("Username is blank, skipping creation of association")
            return False

        if not self.client.get_association(username, account):
            logger.info("Creating association between %s and %s", username, account)
            try:
                self.client.create_association(
                    username, account, self.backend_settings.get("default_account", "root")
                )
            except BackendError as err:
                logger.exception("Unable to create association on backend: %s", err)
                return False
        else:
            logger.info("Association already exists, skipping creation")
        return True

    def remove_users_from_account(self, resource_backend_id: str, usernames: Set[str]) -> List[str]:
        """Remove specified users from the resource on the backend."""
        logger.info(
            "Removing users from account %s on backend: %s",
            resource_backend_id,
            " ,".join(usernames),
        )
        removed_users = []
        for username in usernames:
            try:
                succeeded = self.remove_user(resource_backend_id, username)
                if succeeded:
                    removed_users.append(username)
            except BackendError as e:
                logger.exception(
                    "Unable to remove user %s from account %s, details: %s",
                    username,
                    resource_backend_id,
                    e,
                )
        return removed_users

    def _pre_delete_user_actions(self, account: str, username: str) -> None:
        del account, username

    def remove_user(self, account: str, username: str) -> bool:
        """Delete association between user and an account if it exists."""
        if not account.strip():
            message = "Empty account name"
            raise BackendError(message)

        logger.info("Removing user %s from account %s", username, account)

        if self.client.get_association(username, account):
            logger.info("Deleting association between %s and %s", username, account)
            try:
                self._pre_delete_user_actions(account, username)
                self.client.delete_association(username, account)
            except BackendError as err:
                logger.exception("Unable to delete association in the backend: %s", err)
                return False
        return True

    def set_resource_limits(self, resource_backend_id: str, limits: Dict[str, int]) -> None:
        """Set limits for the resource on the backend."""
        self.client.set_resource_limits(resource_backend_id, limits)

    def get_resource_limits(self, resource_backend_id: str) -> Dict[str, int]:
        """Get limits for the resource on the backend."""
        return self.client.get_resource_limits(resource_backend_id)

    def get_resource_user_limits(self, resource_backend_id: str) -> Dict[str, Dict[str, int]]:
        """Get limits for the resource users on the backend."""
        return self.client.get_resource_user_limits(resource_backend_id)

    def set_resource_user_limits(
        self, resource_backend_id: str, username: str, limits: Dict[str, int]
    ) -> None:
        """Set limits for a specific user in a resource on the backend."""
        logger.info(
            "Setting user %s limits to %s for resource %s", username, limits, resource_backend_id
        )
        self.client.set_resource_user_limits(resource_backend_id, username, limits)

    def _get_allocation_name(self, allocation_slug: str, prefix: str = "") -> str:
        prefix = self.backend_settings.get("allocation_prefix", "")
        return f"{prefix}{allocation_slug}".lower()

    def _get_project_name(self, slug: str) -> str:
        return f"{self.backend_settings.get('project_prefix', '')}{slug}"

    def _get_customer_name(self, slug: str) -> str:
        return f"{self.backend_settings.get('customer_prefix', '')}{slug}"


class UnknownBackend(BaseBackend):
    """Common class for unknown backends."""

    def __init__(self) -> None:
        """Placeholder."""
        super().__init__({}, {})
        self.backend_type = UNKNOWN_BACKEND_TYPE

    def ping(self, _: bool = False) -> bool:
        """Placeholder."""
        return False

    def list_components(self) -> List[str]:
        """Placeholder."""
        return []

    def pull_resources(self, _: List[structures.Resource]) -> Dict[str, structures.Resource]:
        """Placeholder."""
        return {}

    def delete_resource(self, resource_backend_id: str, **kwargs: str) -> None:
        """Placeholder."""
        del kwargs, resource_backend_id

    def create_resource(self, _: Dict) -> structures.Resource:
        """Placeholder."""
        return structures.Resource()

    def downscale_resource(self, account: str) -> bool:
        """Placeholder."""
        del account
        return False

    def pause_resource(self, account: str) -> bool:
        """Placeholder."""
        del account
        return False

    def restore_resource(self, account: str) -> bool:
        """Placeholder."""
        del account
        return False

    def get_resource_metadata(self, _: str) -> dict:
        """Placeholder."""
        return {}

    def add_users_to_resource(
        self, resource_backend_id: str, user_ids: Set[str], **kwargs: dict
    ) -> Set[str]:
        """Placeholder."""
        del kwargs, resource_backend_id
        return user_ids

    def set_resource_limits(self, _: str, limits: Dict[str, int]) -> None:
        """Placeholder."""
        del limits

    def _collect_limits(self, _: Dict[str, Dict]) -> Tuple[Dict[str, int], Dict[str, int]]:
        return {"": 0}, {"": 0}

    def _pull_allocation(self, _: str) -> Optional[structures.Resource]:
        return None

    def _get_usage_report(self, _: List[str]) -> Dict:
        return {}



"""Functions used from external modules."""

import pprint

from waldur_site_agent.backends import logger
from waldur_site_agent.backends.exceptions import BackendError
from waldur_site_agent.backends.slurm_backend.backend import SlurmBackend


def diagnostics(slurm_backend: SlurmBackend) -> bool:
    """Runs diagnostics for SLURM cluster."""
    default_account_name = slurm_backend.backend_settings["default_account"]

    format_string = "{:<30} = {:<10}"
    logger.info(
        format_string.format(
            "SLURM customer prefix", slurm_backend.backend_settings["customer_prefix"]
        )
    )
    logger.info(
        format_string.format(
            "SLURM project prefix", slurm_backend.backend_settings["project_prefix"]
        )
    )
    logger.info(
        format_string.format(
            "SLURM allocation prefix", slurm_backend.backend_settings["allocation_prefix"]
        )
    )
    logger.info(format_string.format("SLURM default account", default_account_name))
    logger.info("")

    logger.info("SLURM tres components:\n%s\n", pprint.pformat(slurm_backend.backend_components))

    try:
        slurm_version_info = slurm_backend.client._execute_command(
            ["-V"], "sinfo", immediate=False, parsable=False
        )
        logger.info("Slurm version: %s", slurm_version_info.strip())
    except BackendError as err:
        logger.error("Unable to fetch SLURM info, reason: %s", err)
        return False

    try:
        slurm_backend.ping(raise_exception=True)
        logger.info("SLURM cluster ping is successful")
    except BackendError as err:
        logger.error("Unable to ping SLURM cluster, reason: %s", err)

    tres = slurm_backend.list_components()
    logger.info("Available tres in the cluster: %s", ",".join(tres))

    default_account = slurm_backend.client.get_account(default_account_name)
    if default_account is None:
        logger.error("There is no account %s in the cluster", default_account)
        return False
    logger.info('Default parent account "%s" is in place', default_account_name)
    logger.info("")

    return True
